/* Module support for GNU Pascal
   Copyright (C) 1994, 95, 96, 97  Free Software Foundation, Inc.

This file is part of GNU Pascal.

GNU Pascal is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU Pascal is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Pascal; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * Authors: Jukka Virtanen <jtv@hut.fi>
 *          Peter Gerwinski <peter.gerwinski@uni-essen.de>
 */

#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "tree.h"
#include "c-tree.h"
#include "input.h"
#include "rtl.h"
#include "obstack.h"
#include "flags.h"

#include "gpc-defs.h"

#ifdef __EMX__
#include <process.h>
#include <stdlib.h>
#include <errno.h>
#endif /* __EMX__ */

#ifdef USG
#define vfork fork
#endif /* USG */

extern tree identifier_output;
extern tree identifier_input;
extern tree global_input_file_node;
extern tree global_output_file_node;

/* Allocates global object for the module list.
 * find_module().
 */
extern struct obstack permanent_obstack;

/* We are currently compiling this module */

module_t module_list = NULL_MODULE;
module_t current_module;

/* A list of all exported names in all modules seen so far.
 *
 * TREE_LIST of exported interfaces we know at the current point of
 * compilation: TREE_VALUE is an IDENTIFIER_NODE of an exported
 * interface name TREE_PURPOSE is a TREE_LIST chain of the names (?
 * decls?) exported by this interface */
tree exported_interface_list;

/* Required module interfaces StandardInput and StandardOutput */
tree standard_interface_input;
tree standard_interface_output;

/* Export "all" mark */
tree export_all;

/* List of the contents of a gpi file */
tree gpi_contents;

/* Flag for debugging the gpi mechanism */
int flag_debug_gpi = 0;

/* Nonzero means do an AutoMake */
int flag_automake = 0;

/* Options to pass to child gpc processes in AutoMake */
char* automake_gpc_options = NULL;

/* Name of a file to store information about targets in */
char* automake_temp_filename = NULL;

/* No comment. :-) */
int we_are_loading_a_gpi_file = 0;


/* Scan the program/module parameter list for entries of FILE_TYPE.
 * If this is at top level, and they are variables of file type,
 * flag the files as external.
 *
 * Since the order of declarations are relaxed, this is checked
 * before every routine.
 *
 * All nodes already handled are marked.
 */
void
associate_external_objects (external_name_list)
     tree external_name_list;
{
  tree link;

  if (external_name_list && top_level_p (current_module->main_program))
    for (link = external_name_list;
	 link;
	 link = TREE_CHAIN (link))
      
      {
	tree id = TREE_VALUE (link);

	if (id == identifier_output)
	  current_module->output_file_node = global_output_file_node;
	else if (id == identifier_input)
	  current_module->input_file_node = global_input_file_node;
	else if (! TREE_PURPOSE (link))
	  {
	    tree name = lookup_name (id);
	    
	    if (name
		&& TREE_CODE (name) == VAR_DECL
		&& TREE_CODE (TREE_TYPE (name)) == FILE_TYPE)
	      {
		PASCAL_EXTERNAL_OBJECT (name) = 1;
		TREE_PURPOSE (link) = error_mark_node;
	      }
	  }
      }
}

/* Check if all the names in program/module param list have been declared
 *
 * If not, give a warning.
 */
void
check_external_objects (idlist)
     tree idlist;
{
  char *what  = current_module->main_program ? "program" : "module";

  for (; idlist; idlist = TREE_CHAIN (idlist))
    {
      tree id = TREE_VALUE (idlist);

      if (id != identifier_output && id != identifier_input)
	{
	  tree name = lookup_name (id);

	  if (name == NULL_TREE)
	    warning ("identifier `%s' in %s heading is undefined",
		     IDENTIFIER_POINTER (id), what);
	  else if (TREE_CODE (name) != VAR_DECL
		   || TREE_CODE (TREE_TYPE (name)) != FILE_TYPE)
	    warning ("identifier `%s' in %s heading is not a variable of file type",
		     IDENTIFIER_POINTER (id), what);
	}
    }
}

/* Possibly add the `static' qualifier.
 */
tree
maybe_make_static (qualifiers)
     tree qualifiers;
{
  int make_static = 1;

  /* @@@@ Need to verify if the function was declared with a
   *      FORWARD directive.
   */
  if (qualifiers)
    {
      tree scan = qualifiers;
      for (; scan; scan = TREE_CHAIN (scan))
	if (TREE_VALUE (scan) == extern_id)
	  {
	    make_static = FALSE;
	    break;
	  }
    }
  
  if (make_static)
    qualifiers = chainon (qualifiers, build_tree_list (NULL_TREE, static_id));

  return qualifiers;
}

/* Locates a module by it's NAME.
 * If CREATE is nonzero, create a new module if old not found
 */
module_t
find_module (name, create)
     tree name;
     int create;
{
  module_t curr;
  
  for (curr = module_list; curr; curr = curr->next)
    {
      if (curr->name == name)
	return curr;
    }
  
  if (create)
    {
      curr = (module_t) obstack_alloc (&permanent_obstack,
				       sizeof (struct module));
      /* Initialize */
      bzero ((void *)curr, sizeof (struct module));
      curr->name  = name;
      curr->next  = module_list;
      module_list = curr;
    }
  return curr;
}

void
initialize_module (id)
     tree id;
{
  current_module = find_module (id, 1);
}

/* Destroy the info about all identifiers in MOD which will
 * be released with the next poplevel().
 */
void
finalize_module (mod)
    struct module *mod;
{
  mod->imports = NULL_TREE;
  mod->exports = NULL_TREE;
  mod->pending_decls = NULL_TREE;
  mod->autoexport = NULL_TREE;
  mod->implementation = NULL_TREE;
  mod->interface = NULL_TREE;
  mod->output_file_node = NULL_TREE;
  mod->input_file_node = NULL_TREE;
  exported_interface_list = NULL_TREE;
}

/* Perform the interface declarations of the current module.
 */
void
handle_pending_decls ()
{
  if (current_module->pending_decls)
    {
      tree scan;
      for (scan = current_module->pending_decls;
	   scan;
	   scan = TREE_CHAIN (scan))
	{
	  if (TREE_PURPOSE (scan) == void_type_node)
	    grok_directive (TREE_VALUE (TREE_VALUE (TREE_VALUE (scan))),
			    TREE_PURPOSE (TREE_VALUE (TREE_VALUE (scan))),
			    TREE_PURPOSE (TREE_VALUE (scan)),
			    0);
	  else if (TREE_PURPOSE (scan) == NULL_TREE)
	    declare_vars (TREE_VALUE (TREE_VALUE (scan)), 	        /* names */
			  TREE_PURPOSE (TREE_VALUE (scan)),	        /* type */
			  TREE_PURPOSE (TREE_CHAIN (TREE_VALUE (scan))),/* init */
			  TREE_VALUE (TREE_CHAIN (TREE_VALUE (scan))),  /* qual */
			  0);
	  else
	    abort ();
	}
      current_module->pending_decls = NULL_TREE;
    }
}

/* NAME is an IDENTIFIER_NODE of the exported interface name,
 * or a TREE_LIST if exporting a range (TREE_PURPOSE..TREE_VALUE).
 * EXPORT_LIST:
 *   TREE_LIST
 *       TREE_PURPOSE: Export renaming (new name) or NULL_TREE.
 *       TREE_VALUE  : IDENTIFIER_NODE of the exported name
 */
void
export_interface (name, export_list)
     tree name;
     tree export_list;
{
  tree exported;
  tree exported_names = NULL_TREE;
  tree nscan, new_export_list;

  for (exported = exported_interface_list;
       exported; exported = TREE_CHAIN (exported))
    if (TREE_VALUE (exported) == name)
      {
	error ("interface `%s' has already been exported",
	       IDENTIFIER_POINTER (name));
	return;
      }

  if (export_list == export_all)
    exported_names = build_tree_list (NULL_TREE, NULL_TREE);
  else
    for (nscan = export_list; nscan; nscan = TREE_CHAIN (nscan))
      {
        tree value = build_tree_list (TREE_PURPOSE (nscan), TREE_VALUE (nscan));
        exported_names = chainon (exported_names, value);
      }
  
  current_module->exports
    = chainon (current_module->exports,
	       build_tree_list (exported_names, name));

  new_export_list = build_tree_list (exported_names, name);
  exported_interface_list = chainon (exported_interface_list, new_export_list);
  if (export_list == export_all)
    current_module->autoexport = chainon (current_module->autoexport,
                                          build_tree_list (NULL_TREE,
                                                           new_export_list));
}

int
name_exported_p (name)
     tree name;
{
  tree scan;
  for (scan = current_module->exports; scan; scan = TREE_CHAIN (scan))
    {
      tree id_chain;

      for (id_chain = TREE_PURPOSE (scan); id_chain;
           id_chain = TREE_CHAIN (id_chain))
        {
          /* Exported ranges are not taken into account here
           * because name_exported_p() is never called for them.
           */
          if (name == TREE_VALUE (id_chain))
            return 1;
        }
    }
  return 0;
}

/*
 * Export names of a Borland Pascal Unit.
 * This also handles "export foo = all" clauses
 * and perhaps, one day, PXSC modules.
 */
void
handle_autoexport (name)
     tree name;
{
  tree e;
  if (! we_are_loading_a_gpi_file)
    for (e = current_module->autoexport; e; e = TREE_CHAIN (e))
      {
        tree nscan, p = TREE_PURPOSE (TREE_VALUE (e));
        for (nscan = p; nscan; nscan = TREE_CHAIN (nscan))
          {
            if (TREE_VALUE (nscan) == name)
              break;
          }
        if (!nscan)
          p = chainon (p, module_export_clause (name, NULL_TREE, 0));
      }
}



/*
 *  The AutoMake facility
 *
 *  When invoked with an --automake option, gpc can do an
 *  implicit `make' for preprocessing, compiling and linking.
 *  A Makefile is not needed; everything is extracted from
 *  the Pascal source.
 *
 *  Read the `AutoMake' section of the Texinfo documentation
 *  for more information about this.
 *
 *  - PG
 */

/*
 * For the AutoMake facility we need to execute gpc as a child
 * process.  We do it here by copying pexecute() from gcc.c to
 * make it easier to keep track of modifications in gcc, and by
 * providing an easier execute_child() interface function which
 * does not use pipes.
 */

/*
 * Definitions from gcc.c
 */

extern int execv (), execvp ();

/* If a stage of compilation returns an exit status >= 1,
   compilation of that file ceases.  */

#define MIN_FATAL_STATUS 1

#ifndef WIFSIGNALED
#define WIFSIGNALED(S) (((S) & 0xff) != 0 && ((S) & 0xff) != 0x7f)
#endif

#ifndef WIFEXITED
#define WIFEXITED(S) (((S) & 0xff) == 0)
#endif

#ifndef WEXITSTATUS
#define WEXITSTATUS(S) (((S) & 0xff00) >> 8)
#endif

/*
 * Provide a function to report an error if a child's execution fails
 */
static void
perror_exec (name)
     char *name;
{
  error ("installation problem, cannot exec `%s'", name);
}

#ifdef __EMX__
static int pipes_supported = 0;
#endif /* __EMX__ */

/* Unmodified copied part from gcc.c starts here. */

/* stdin file number.  */
#define STDIN_FILE_NO 0

/* stdout file number. */
#define STDOUT_FILE_NO 1

/* value of `pipe': port index for reading.  */
#define READ_PORT 0

/* value of `pipe': port index for writing.  */
#define WRITE_PORT 1

/* Pipe waiting from last process, to be used as input for the next one.
   Value is STDIN_FILE_NO if no pipe is waiting
   (i.e. the next command is the first of a group).  */

static int last_pipe_input;

/* Fork one piped subcommand.  FUNC is the system call to use
   (either execv or execvp).  ARGV is the arg vector to use.
   NOT_LAST is nonzero if this is not the last subcommand
   (i.e. its output should be piped to the next one.)  */

#ifdef __MSDOS__

#include <process.h>
static int
pexecute (search_flag, program, argv, not_last)
     int search_flag;
     char *program;
     char *argv[];
     int not_last;
{
#ifdef __GO32__
  int i = (search_flag ? spawnv : spawnvp) (1, program, argv);
#else
  char *scmd, *rf;
  FILE *argfile;
  int i, el = search_flag ? 0 : 4;

  scmd = (char *)malloc (strlen (program) + strlen (temp_filename) + 6 + el);
  rf = scmd + strlen(program) + 2 + el;
  sprintf (scmd, "%s%s @%s.gp", program,
	   (search_flag ? "" : ".exe"), temp_filename);
  argfile = fopen (rf, "w");
  if (argfile == 0)
    pfatal_with_name (rf);

  for (i=1; argv[i]; i++)
    {
      char *cp;
      for (cp = argv[i]; *cp; cp++)
	{
	  if (*cp == '"' || *cp == '\'' || *cp == '\\' || isspace (*cp))
	    fputc ('\\', argfile);
	  fputc (*cp, argfile);
	}
      fputc ('\n', argfile);
    }
  fclose (argfile);

  i = system (scmd);

  remove (rf);
#endif
  
  if (i == -1)
    {
      perror_exec (program);
      return MIN_FATAL_STATUS << 8;
    }
  return i << 8;
}

#endif

/* include for Unix-like environments but not for Dos-like environments */
#if !defined(__MSDOS__) && !defined(OS2) &&(defined(__CYGWIN32__) || !defined(_WIN32))

#ifdef __EMX__

static int
pexecute (search_flag, program, argv, not_last)
     int search_flag;
     char *program;
     char *argv[];
     int not_last;
{
  int pid;
  int pdes[2], org_stdin, org_stdout;
  int input_desc = last_pipe_input;
  int output_desc = STDOUT_FILE_NO;

  /* If this isn't the last process, make a pipe for its output,
     and record it as waiting to be the input to the next process.  */

  if (not_last)
    {
      if (pipe (pdes) < 0)
	pfatal_with_name ("pipe");
      output_desc = pdes[WRITE_PORT];
      last_pipe_input = pdes[READ_PORT];
    }
  else
    last_pipe_input = STDIN_FILE_NO;

  if (pipes_supported && input_desc != STDIN_FILE_NO)
    {
      org_stdin = dup (STDIN_FILE_NO);
      dup2 (input_desc, STDIN_FILE_NO);
      close (input_desc); 
    }
  if (pipes_supported && output_desc != STDOUT_FILE_NO)
    {
      org_stdout = dup (STDOUT_FILE_NO);
      dup2 (output_desc, STDOUT_FILE_NO);
      close (output_desc);
    }
  pid = (search_flag ? spawnv : spawnvp) (P_NOWAIT, program,
                                          (char const * const *)argv);
  if (pipes_supported && input_desc != STDIN_FILE_NO)
    {
      dup2 (org_stdin, STDIN_FILE_NO);
      close (org_stdin);
    }
  if (pipes_supported && output_desc != STDOUT_FILE_NO)
    {
      dup2 (org_stdout, STDOUT_FILE_NO);
      close (org_stdout);
    }

  if (pid == -1)
    {
      perror_exec (program);
      exit (-1);
    }

  return pid;
}

#else /* not __EMX__ */

static int
pexecute (search_flag, program, argv, not_last)
     int search_flag;
     char *program;
     char *argv[];
     int not_last;
{
  int (*func)() = (search_flag ? execv : execvp);
  int pid;
  int pdes[2];
  int input_desc = last_pipe_input;
  int output_desc = STDOUT_FILE_NO;
  int retries, sleep_interval;

  /* If this isn't the last process, make a pipe for its output,
     and record it as waiting to be the input to the next process.  */

  if (not_last)
    {
      if (pipe (pdes) < 0)
	pfatal_with_name ("pipe");
      output_desc = pdes[WRITE_PORT];
      last_pipe_input = pdes[READ_PORT];
    }
  else
    last_pipe_input = STDIN_FILE_NO;

  /* Fork a subprocess; wait and retry if it fails.  */
  sleep_interval = 1;
  for (retries = 0; retries < 4; retries++)
    {
      pid = vfork ();
      if (pid >= 0)
	break;
      sleep (sleep_interval);
      sleep_interval *= 2;
    }

  switch (pid)
    {
    case -1:
#ifdef vfork
      pfatal_with_name ("fork");
#else
      pfatal_with_name ("vfork");
#endif
      /* NOTREACHED */
      return 0;

    case 0: /* child */
      /* Move the input and output pipes into place, if nec.  */
      if (input_desc != STDIN_FILE_NO)
	{
	  close (STDIN_FILE_NO);
	  dup (input_desc);
	  close (input_desc);
	}
      if (output_desc != STDOUT_FILE_NO)
	{
	  close (STDOUT_FILE_NO);
	  dup (output_desc);
	  close (output_desc);
	}

      /* Close the parent's descs that aren't wanted here.  */
      if (last_pipe_input != STDIN_FILE_NO)
	close (last_pipe_input);

      /* Exec the program.  */
      (*func) (program, argv);
      perror_exec (program);
      exit (-1);
      /* NOTREACHED */
      return 0;

    default:
      /* In the parent, after forking.
	 Close the descriptors that we made for this child.  */
      if (input_desc != STDIN_FILE_NO)
	close (input_desc);
      if (output_desc != STDOUT_FILE_NO)
	close (output_desc);

      /* Return child's process number.  */
      return pid;
    }
}

#endif /* not __EMX__ */
#endif /* ! __MSDOS__ && ! OS2 && && (__CYGWIN32___ || ! _WIN32) */

#if defined(OS2)

static int
pexecute (search_flag, program, argv, not_last)
     int search_flag;
     char *program;
     char *argv[];
     int not_last;
{
  return (search_flag ? spawnv : spawnvp) (1, program, argv);
}
#endif /* OS2 */

#if defined (_WIN32) && !defined (__CYGWIN32__)

#include <process.h>
/* ??? Why are these __spawnv{,p} and not _spawnv{,p}?  */
extern int __spawnv ();
extern int __spawnvp ();

static int
pexecute (search_flag, program, argv, not_last)
     int search_flag;
     char *program;
     char *argv[];
     int not_last;
{
  return (search_flag ? __spawnv : __spawnvp) (1, program, argv);
}
#endif /* _WIN32 */

/*
 * Unmodified copied part from gcc.c ends here.
 */

/*
 * The following execute_child() function is inspired by
 * execute() of gcc.c, so keep track of changes there!
 */
static int
execute_child (argv)
     char *argv[];
{
  int status;
  int pid;
  last_pipe_input = STDIN_FILE_NO;
  if (version_flag)  /* i.e. gpc was invoked with --verbose */
    {
      char **j;
      fprintf (stderr, "GPC AutoMake: ");
      for (j = argv; *j; j++)
        fprintf (stderr, " %s", *j);
      fprintf (stderr, "\n");
    }
  pid = pexecute (0, argv [0], argv, 0);
#ifdef __MSDOS__
  status = pid;
#else
#if defined (_WIN32) && !defined (__CYGWIN32__)
  pid = cwait (&status, pid, WAIT_CHILD);
#else
  pid = wait (&status);
#endif
#endif
#ifdef __EMX__
  while (pid < 0 && errno == EINTR)
    pid = wait (&status);
#endif /* __EMX__ */
  if (pid < 0)
    abort ();
  if (WIFSIGNALED (status)
      || WIFEXITED (status) && WEXITSTATUS (status) >= MIN_FATAL_STATUS)
    {
      /* An error message already has been put out by the child gpc process. */
      return -1;
    }
  else
    return 0;
}

/*
 * Add an object file to the list in the AutoMake temporary file.
 * The contents of that file will be passed to the linker.
 */
void
add_to_automake_temp_file (object_filename)
     char *object_filename;
{
  if (automake_temp_filename && object_filename)
    {
       FILE *automake_temp_file = fopen (automake_temp_filename, "at");
       if (automake_temp_file)
         {
           fprintf (automake_temp_file, "%s\n", object_filename);
           fclose (automake_temp_file);
         }
    }
}

/*
 * Store the name of the executable to be produced in
 * the AutoMake temporary file.
 */
void
store_executable_name ()
{
  if (executable_file_name)
    {
      /* Store the name of the program in the automake temp
       * file, so the executable can be named after it.
       */
      char *name;
      if (*executable_file_name)
        ; /* accept it */
      else if (main_input_filename)
        {
	  char *p;
	  executable_file_name = xmalloc (strlen (main_input_filename));
	  strcpy (executable_file_name, main_input_filename);
	  p = executable_file_name + strlen (executable_file_name) - 1;
	  while (p > executable_file_name && *p != '.')
	    p--;
	  if (*p == '.' )
	    *p = 0;
	}
      else  /* Should not happen */
        executable_file_name = "a.out";
      name = xmalloc (strlen (executable_file_name) + 8);
      strcpy (name, "-o");
      strcat (name, executable_file_name);
#if defined (__MSDOS__) || defined (__EMX__) || defined (__CYGWIN32__)
      strcat (executable_file_name, ".exe");
#endif
      add_to_automake_temp_file (name);
      free (name);
    }
}

/*
 * Pass the AutoMake options to a child process
 */
int pass_automake_gpc_options (argv, i)
     char **argv;
     int i;
{
  char recursive_options [] = "--automake";
  if (automake_gpc_options)
    {
      /* Pass the options specified behind --automake */
      char *p = save_string (automake_gpc_options);
      argv [i++] = p;
      while (*p)
        {
          if (*p == ' ')
            {
              *p = 0;
              p++;
              argv [i++] = p;
            }
          else
            p++;
        } 
      /* Pass the --automake specification itself */
      p = concat ( "=\"", automake_gpc_options, "\"");
      argv [i++] = concat (recursive_options, p, "");
    }
  else
    argv [i++] = save_string (recursive_options);
  if (automake_temp_filename)
    argv [i++] = concat ("--amtmpfile=", automake_temp_filename, "");
  return i;
}

/*
 * Check whether a module must be recompiled (for AutoMake).
 */
int
module_must_be_recompiled (gpi_filename, module_filename)
     char *gpi_filename;
     char *module_filename;
{
  struct stat gpi_status, source_status;
  char *argv [256], *dep_filename, *p;
  int i;
  FILE *dep_file;

  /* Module is in this file --> cannot recompile */
  if (!strcmp (main_input_filename, module_filename))
    return 0;

  /* no source found --> cannot recompile */
  if (access (module_filename, 4) != 0)
    return 0;

  /* gpi older than source --> recompile
   */
  stat (gpi_filename, &gpi_status);
  stat (module_filename, &source_status);
  if (gpi_status.st_mtime < source_status.st_mtime)
    return 1;

  /* Run the preprocessor with -M for creating a dependency file.
   * The name of that file is derived from module_filename by 
   * replacing `.p' or `.pas' with `.d' (like gcc does with the 
   * -MD option).
   *
   * @@@ Should do this when compiling the module and store the
   * information in the GPI file.  That would be faster.
   */
  dep_filename = xmalloc (strlen (module_filename) + 3);
  strcpy (dep_filename, module_filename);
  p = dep_filename + strlen (dep_filename);
  do
    p--;
  while (p >= dep_filename && *p != '.');
  *p = 0;
  strcat (dep_filename, ".d");
  i = 0;
  argv [i++] = "gpc";
  i = pass_automake_gpc_options (argv, i);
  argv [i++] = "-M";
  argv [i++] = module_filename;
  argv [i++] = "-o";
  argv [i++] = dep_filename;
  argv [i++] = NULL;
  execute_child (argv);

  /* Parse the `.d' file and check if the object file or the gpi file
   * is older than one of the source files.  If yes --> recompile.
   */
  dep_file = fopen (dep_filename, "rt");
  if (dep_file)
    {
      char source_filename [256], *object_filename = NULL;
      while (! feof (dep_file))
        {
          fscanf (dep_file, "%s", source_filename);
          if (strcmp (source_filename, "\\") != 0)
            {
              p = source_filename + strlen (source_filename) - 1;
              if (*p == ':')
                {
                  /* This is the name of the object file.
                   * If it doesn't exist, we must compile.
                   * If it is older than the gpi file,
                   * update the time.
                   */
                  *p = 0;
                  object_filename = save_string (source_filename);
                  if (access (object_filename, 4) != 0)
                    {
                      fclose (dep_file);
                      unlink (dep_filename);
                      return 1;
                    }
                  stat (object_filename, &source_status);
                  if (gpi_status.st_mtime > source_status.st_mtime)
                    gpi_status.st_mtime = source_status.st_mtime;
                }
              else
                {
                  /* This is the name of a source file.
                   * If it is younger than the gpi file,
                   * recompile.
                   */
                  stat (source_filename, &source_status);
                  if (gpi_status.st_mtime < source_status.st_mtime)
                    {
                      fclose (dep_file);
                      unlink (dep_filename);
                      return 1;
                    }
                }
            }
        }
      fclose (dep_file);
      unlink (dep_filename);
      /* Recompile not needed.  Tell the linker about the object file. */
      if (object_filename)
        add_to_automake_temp_file (object_filename);
    }
  /* Otherwise: no further check possible */

  /* All possible checks passed --> no recompilation.
   */
  return 0;
}

/*
 * Compile a module during an AutoMake.
 */
int
compile_module (filename)
     char *filename;
{
  char *argv [256];
  if (access (filename, 4) == 0)
    {
      int result, i = 0;
      char *object_filename, *p;
 
      /* Tell the linker about the object file */
      object_filename = xmalloc (strlen (filename) + 3);
      strcpy (object_filename, filename);
      p = object_filename + strlen (filename) - 1;
      while (p > object_filename && *p != '.')
        p--;
      p++;
      *p = 0;
      strcat (object_filename, "o");
      add_to_automake_temp_file (object_filename);

      argv [i++] = "gpc";

      /* Pass automake GPC options only if Pascal source.
       */
      p = filename + strlen (filename) - 1;
      while (p > filename && *p != '.')
        p--;
      p++;
      if (strcmp (p, "p") == 0 || strcmp (p, "pas") == 0)
        i = pass_automake_gpc_options (argv, i);

      argv [i++] = "-c";
      argv [i++] = filename;
      argv [i++] = NULL;
      result = execute_child (argv);
      free (object_filename);
      return result;
    }
  else
    return -1;
}



/*
 *  GPI file handling -- storing and retreiving tree nodes in an
 *  implementation-dependend (but not *too* implementation-dependend ;-)
 *  "GNU Pascal Interface" binary file.
 *
 *  For an introduction, read the `GPI' section of the Texinfo documentation.
 *
 *  - PG
 */

/* Store a string in a binary stream
 */
void
store_string (f, s)
     FILE *f;
     const char *s;
{
  unsigned char l = strlen (s);
  fwrite (&l, 1, 1, f);
  if (l > 0)
    fwrite (s, l, 1, f);
}

/* Load a string out of a binary stream
 */
void
load_string (f, s)
     FILE *f;
     char *s;
{
  unsigned char l;
  fread (&l, sizeof (unsigned char), 1, f);
  if (l > 0)
    fread (s, l, 1, f);
  s [l] = 0;
}

/* Store a tree node's flags in a stream
 */
void
store_flags (t, f)
     tree t;
     FILE *f;
{
  fwrite ((tree *) t + 2, 4, 1, f);
}

/* Load a tree node's flags out of a stream
 */
void
load_flags (t, f)
     tree t;
     FILE *f;
{
  fread ((tree *) t + 2, 4, 1, f);
}

/* Store a *_TYPE tree node's extra flags in a stream
 */
void
store_type_flags (t, f)
     tree t;
     FILE *f;
{
  fwrite (&TYPE_UID (t), sizeof (unsigned) + 4 + sizeof (int), 1, f);
}

/* Load a *_TYPE tree node's extra flags out of a stream
 */
void
load_type_flags (t, f)
     tree t;
     FILE *f;
{
  fread (&TYPE_UID (t), sizeof (unsigned) + 4 + sizeof (int), 1, f);
}

/* Store a *_DECL tree node's extra flags in a stream
 */
void
store_decl_flags (t, f)
     tree t;
     FILE *f;
{
  fwrite (&DECL_UID (t), sizeof (unsigned) + 4, 1, f);
  fwrite (&DECL_FRAME_SIZE (t), sizeof (int), 1, f);
}

/* Load a *_DECL tree node's extra flags out of a stream
 */
void
load_decl_flags (t, f)
     tree t;
     FILE *f;
{
  fread (&DECL_UID (t), sizeof (unsigned) + 4, 1, f);
  fread (&DECL_FRAME_SIZE (t), sizeof (int), 1, f);
}

/*
 * Store a tree (containing a declaration) in a stream.
 * (A generalized version of this should go into tree.c.)
 */
void
store_tree (t, s, depth)
     tree t;
     FILE *s;
     int depth;
{
  unsigned char code;

  if (!t)
    {
      if (flag_debug_gpi)
        fprintf (stderr, "storing (%d): NULL_TREE\n", depth);
      code = (unsigned char) LAST_AND_UNUSED_TREE_CODE;
      fwrite (&code, 1, 1, s);
      return;
    }

  #define PROCESS_NODE(n,node)                                \
  if (t == node)                                              \
    {                                                         \
      code = (unsigned char) LAST_AND_UNUSED_TREE_CODE + n;   \
      fwrite (&code, 1, 1, s);                                \
      if (flag_debug_gpi)                                     \
        {                                                     \
          char msg[] = GPC_MAKE_STRING (node);                \
          fprintf (stderr, "storing (%d): %s\n", depth, msg); \
        }                                                     \
      return;                                                 \
    }
  #include "gpc-nodes.def"
  #undef PROCESS_NODE

  if (depth > 0)
    {
      tree c;
      int count = 0;
      for (c = gpi_contents; c; c = TREE_CHAIN (c))
        {
          if (t == TREE_VALUE (c))
            {
              if (flag_debug_gpi)
                {
                  fprintf (stderr, "storing (%d, via gpi_contents):\n", depth);
                  debug_tree (t);
                }
              code = 0xFF;
              fwrite (&code, 1, 1, s);
              fwrite (&count, sizeof (int), 1, s);
              return;
            }
          count++;
        }
    }
  if (flag_debug_gpi)
    {
      fprintf (stderr, "storing (%d):\n", depth);
      debug_tree (t);
    }
  code = (unsigned char) TREE_CODE (t);
  fwrite (&code, 1, 1, s);
  switch (TREE_CODE (t))
    {
      #define STORE_MAIN_VARIANT(t)                                \
              if (TYPE_MAIN_VARIANT (t) == t)                      \
                store_tree (error_mark_node, s, depth + 1);        \
              else                                                 \
                store_tree (TYPE_MAIN_VARIANT (t), s, depth + 1);  \
                
      case IDENTIFIER_NODE:
        {
          store_string (s, IDENTIFIER_POINTER (t));
          store_flags (t, s);
          if (depth == 0)
            {
              if (IDENTIFIER_LOCAL_VALUE (t))
                store_tree (IDENTIFIER_LOCAL_VALUE (t), s, depth + 1);
              else if (IDENTIFIER_GLOBAL_VALUE (t))
                store_tree (IDENTIFIER_GLOBAL_VALUE (t), s, depth + 1);
              else
                store_tree (IDENTIFIER_LIMBO_VALUE (t), s, depth + 1);
            }
          break;
        }

      case TREE_LIST:
        {
          store_tree (TREE_VALUE (t), s, depth + 1);
          store_tree (TREE_CHAIN (t), s, depth + 1);
          store_flags (t, s);
          break;
        }

      case VOID_TYPE:
      case REAL_TYPE:
      case COMPLEX_TYPE:
      case ENUMERAL_TYPE:
      case BOOLEAN_TYPE:
      case CHAR_TYPE:
      case SET_TYPE:
      case LANG_TYPE:
        {
          store_flags (t, s);
          store_type_flags (t, s);
          store_tree (TYPE_SIZE (t), s, depth + 1);
          STORE_MAIN_VARIANT (t);
          break;
        }

      case INTEGER_TYPE:
        {
          store_flags (t, s);
          store_type_flags (t, s);
          store_tree (TYPE_SIZE (t), s, depth + 1);
          store_tree (TYPE_MIN_VALUE (t), s, depth + 1);
          store_tree (TYPE_MAX_VALUE (t), s, depth + 1);
          STORE_MAIN_VARIANT (t);
          break;
        }
      
      case POINTER_TYPE:
      case REFERENCE_TYPE:
        {
          store_flags (t, s);
          store_type_flags (t, s);
          store_tree (TYPE_SIZE (t), s, depth + 1);
          gpi_contents = chainon (gpi_contents, build_tree_list (NULL_TREE, t));
          store_tree (TREE_TYPE (t), s, depth + 1);
          STORE_MAIN_VARIANT (t);
          return;
          /* Pointers and references might cause infinite */
          /* loops in the gpi file otherwise.             */
        }

      case FILE_TYPE:
        {
          /* Since FILE_TYPE is not documented in tree.def, */
          /* I only can guess about its structure.          */
          store_flags (t, s);
          store_type_flags (t, s);
          store_tree (TREE_TYPE (t), s, depth + 1);
          STORE_MAIN_VARIANT (t);
          break;
        }

      case ARRAY_TYPE:
        {
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_tree (TYPE_DOMAIN (t), s, depth + 1);
          store_flags (t, s);
          store_type_flags (t, s);
          STORE_MAIN_VARIANT (t);
          break;
        }

      case RECORD_TYPE:
      case UNION_TYPE:
        {
          tree f;
          char lang_code = 0;
          store_flags (t, s);
          store_type_flags (t, s);
          store_tree (TYPE_SIZE (t), s, depth + 1);
          if (TYPE_LANG_SPECIFIC (t))
            {
              lang_code = TYPE_LANG_CODE (t);
              fwrite (&lang_code, 1, 1, s);
              store_tree (TYPE_LANG_NAME (t), s, depth + 1);
              store_tree (TYPE_LANG_BASE (t), s, depth + 1);
            }
          else
            fwrite (&lang_code, 1, 1, s);
          gpi_contents = chainon (gpi_contents, build_tree_list (NULL_TREE, t));
          for (f = TYPE_FIELDS (t); f; f = TREE_CHAIN (f))
            {
              store_tree (f, s, depth + 1);
              store_tree (DECL_FIELD_BITPOS (f), s, depth + 1);
            }
          store_tree (NULL_TREE, s, depth + 1);  /* end of field decls */
          STORE_MAIN_VARIANT (t);
          return;
          /* To save disk space, cut here too -- not */
          /* only at POINTER_TYPE and REFRENCE_TYPE. */
        }

      case FUNCTION_TYPE:
        {
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_tree (TYPE_ARG_TYPES (t), s, depth + 1);
          store_flags (t, s);
          store_type_flags (t, s);
          break;
        }

      case INTEGER_CST:
        {
          fwrite (&TREE_INT_CST_LOW (t), sizeof (HOST_WIDE_INT), 1, s);
          fwrite (&TREE_INT_CST_HIGH (t), sizeof (HOST_WIDE_INT), 1, s);
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_flags (t, s);
          break;
        }

      case REAL_CST:
        {
          store_flags (t, s);
          fwrite (&TREE_REAL_CST (t), sizeof (REAL_VALUE_TYPE), 1, s);
          store_tree (TREE_TYPE (t), s, depth + 1);
          break;
        }

      case COMPLEX_CST:
        {
          store_flags (t, s);
          store_tree (TREE_REALPART (t), s, depth + 1);
          store_tree (TREE_IMAGPART (t), s, depth + 1);
          store_tree (TREE_TYPE (t), s, depth + 1);
          break;
        }

      case STRING_CST:
        {
          store_flags (t, s);
          fwrite (&TREE_STRING_LENGTH (t), sizeof (int), 1, s);
          fwrite (TREE_STRING_POINTER (t), TREE_STRING_LENGTH (t) + 1, 1, s );
          store_tree (TREE_TYPE (t), s, depth + 1);
          break;
        }

      case FUNCTION_DECL:
        {
          store_tree (DECL_NAME (t), s, depth + 1);  /* name of function */
          store_tree (DECL_ASSEMBLER_NAME (t), s, depth + 1);
          store_tree (TREE_TYPE (TREE_TYPE (t)), s, depth + 1);  /* return type */
          store_tree (TYPE_ARG_TYPES (TREE_TYPE (t)), s, depth + 1);  /* argument types */
          store_flags (t, s);  /* This might be a constructor or a virtual method. */
          break;
        }

      case LABEL_DECL:
      case PARM_DECL:
      case RESULT_DECL:
        {
          store_tree (DECL_NAME (t), s, depth + 1);  /* name of function */
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_flags (t, s);
          break;
        }

      case FIELD_DECL:
        {
          store_flags (t, s);
          store_decl_flags (t, s);
          store_tree (DECL_NAME (t), s, depth + 1);
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_tree (DECL_SIZE (t), s, depth + 1);
          break;
        }

      case CONST_DECL:
      case TYPE_DECL:
        {
          store_flags (t, s);
          store_tree (DECL_NAME (t), s, depth + 1);
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_tree (DECL_INITIAL (t), s, depth + 1);
          break;
        }

      case VAR_DECL:
        {
          store_tree (DECL_NAME (t), s, depth + 1);
          store_tree (DECL_ASSEMBLER_NAME (t), s, depth + 1);
          store_tree (TREE_TYPE (t), s, depth + 1);
          store_flags (t, s);
          /* don't store DECL_INITIAL */
          break;
        }
      #undef STORE_MAIN_VARIANT
    }
  gpi_contents = chainon (gpi_contents, build_tree_list (NULL_TREE, t));
}

/*
 * Load (parts of) a tree (containing a declaration) out of a stream.
 * (A generalized version of this should go into tree.c.)
 */
tree
load_tree (s, depth)
     FILE *s;
     int depth;
{
  unsigned char code;
  tree t = NULL_TREE;

  fread (&code, 1, 1, s);
  if (code == LAST_AND_UNUSED_TREE_CODE)
    {
      if (flag_debug_gpi)
        fprintf (stderr, "loaded (%d): NULL_TREE\n", depth);
      return NULL_TREE;
    }
  push_obstacks_nochange ();
  end_temporary_allocation ();
  switch (code)
    {
      #define PROCESS_NODE(n,node)                                \
      case LAST_AND_UNUSED_TREE_CODE + n:                         \
        {                                                         \
          t = node;                                               \
          if (flag_debug_gpi)                                     \
            {                                                     \
              char msg[] = GPC_MAKE_STRING (node);                \
              fprintf (stderr, "loaded (%d): %s\n", depth, msg);  \
            }                                                     \
          return t;                                               \
        }
      #include "gpc-nodes.def"
      #undef PROCESS_NODE

      #define LOAD_MAIN_VARIANT(t)                                 \
              TYPE_MAIN_VARIANT (t) = load_tree (s, depth + 1);    \
              if (TYPE_MAIN_VARIANT (t) == error_mark_node)        \
                TYPE_MAIN_VARIANT (t) = t;

      case 0xFF:
        {
          tree c = gpi_contents;
          int count;
          fread (&count, sizeof (int), 1, s);
          while (count-- && c)
            c = TREE_CHAIN (c);
          if (!c)
            {
              error ("invalid gpi file");
              t = error_mark_node;
            }
          else
            t = TREE_VALUE (c);

          if (flag_debug_gpi)
            {
              fprintf (stderr, "loaded (%d, via gpi_contents):\n", depth);
              debug_tree (t);
            }

          return t;
        }

      case IDENTIFIER_NODE:
        {
          char id [256];
          load_string (s, id);
          t = get_identifier (id);
          load_flags (t, s);
          if (depth == 0)
            {
              IDENTIFIER_LIMBO_VALUE (t) = load_tree (s, depth + 1);
              IDENTIFIER_LOCAL_VALUE (t) = NULL_TREE;  /* for renaming etc. */
            }
          break;
        }

      case TREE_LIST:
        {
          tree value, chain;
          value = load_tree (s, depth + 1);
          chain = load_tree (s, depth + 1);
          t = chainon (build_tree_list (NULL_TREE, value), chain);
          load_flags (t, s);
          break;
        }

      case VOID_TYPE:
      case REAL_TYPE:
      case COMPLEX_TYPE:
      case ENUMERAL_TYPE:
      case BOOLEAN_TYPE:
      case CHAR_TYPE:
      case SET_TYPE:
      case LANG_TYPE:
        {
          t = make_node (code);
          load_flags (t, s);
          load_type_flags (t, s);
          TYPE_SIZE (t) = load_tree (s, depth + 1);
          LOAD_MAIN_VARIANT (t);
          break;
        }

      case INTEGER_TYPE:
        {
          t = make_node (code);
          load_flags (t, s);
          load_type_flags (t, s);
          TYPE_SIZE (t) = load_tree (s, depth + 1);
          TYPE_MIN_VALUE (t) = load_tree (s, depth + 1);
          TYPE_MAX_VALUE (t) = load_tree (s, depth + 1);
          LOAD_MAIN_VARIANT (t);
          break;
        }
      
      case POINTER_TYPE:
      case REFERENCE_TYPE:
        {
          t = make_node (code);
          load_flags (t, s);
          load_type_flags (t, s);
          TYPE_SIZE (t) = load_tree (s, depth + 1);
          gpi_contents = chainon (gpi_contents, build_tree_list (NULL_TREE, t));
          TREE_TYPE (t) = load_tree (s, depth + 1);
          LOAD_MAIN_VARIANT (t);
          return t;
          /* Pointers and references might cause infinite */
          /* loops in the gpi file otherwise.             */
        }

      case FILE_TYPE:
        {
          /* Since FILE_TYPE is not documented in tree.def, */
          /* I only can guess about its structure.          */
          t = make_node (code);
          load_flags (t, s);
          load_type_flags (t, s);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          LOAD_MAIN_VARIANT (t);
          break;
        }

      case ARRAY_TYPE:
        {
          tree type, domain;
          type = load_tree (s, depth + 1);
          domain = load_tree (s, depth + 1);
          /* In case this array already exists, we do not want to
           * spoil its flags, thus we copy_node it.
           */
          t = copy_node (build_array_type (type, domain));
          load_flags (t, s);
          load_type_flags (t, s);
          LOAD_MAIN_VARIANT (t);
          break;
        }

      case RECORD_TYPE:
      case UNION_TYPE:
        {
          tree f, f0 = NULL_TREE;
          tree current_type_name_save = current_type_name;
          char lang_code;
          t = make_node (code);
          load_flags (t, s);
          load_type_flags (t, s);
          TYPE_SIZE (t) = load_tree (s, depth + 1);
          fread (&lang_code, 1, 1, s);
          if (lang_code)
            {
              TYPE_LANG_SPECIFIC (t) = allocate_type_lang_specific ();
              TYPE_LANG_CODE (t) = lang_code;
              TYPE_LANG_NAME (t) = load_tree (s, depth + 1);
              TYPE_LANG_BASE (t) = load_tree (s, depth + 1);
              if (lang_code == PASCAL_LANG_OBJECT)
                current_type_name = TYPE_LANG_NAME (t);
            }
          gpi_contents = chainon (gpi_contents, build_tree_list (NULL_TREE, t));
          do
            {
              f = load_tree (s, depth + 1);
              if (f)
                {
                  /* Don't use chainon: danger of circularity */
                  /* if t0 was loaded via gpi_contents.       */
                  if (f0)
                    TREE_CHAIN (f0) = f;
                  else
                    TYPE_FIELDS (t) = f;
                  /* Delete possibly invalid TREE_CHAIN info  */
                  /* if t was loaded via gpi_contents.        */
                  TREE_CHAIN (f) = NULL_TREE;
                  f0 = f;
                  DECL_CONTEXT (f) = t;
                  DECL_FIELD_BITPOS (f) = load_tree (s, depth + 1);
                }
            }
          while (f);
          LOAD_MAIN_VARIANT (t);
          current_type_name = current_type_name_save;  /* usually NULL_TREE */
          return t;
          /* To save disk space, cut here too -- not */
          /* only at POINTER_TYPE and REFRENCE_TYPE. */
        }

      case FUNCTION_TYPE:
        {
          tree type = load_tree (s, depth + 1);
          tree arg_types = load_tree (s, depth + 1);
          t = build_function_type (type, arg_types);
          load_flags (t, s);
          load_type_flags (t, s);
          break;
        }

      case INTEGER_CST:
        {
          t = make_node (code);
          fread (&TREE_INT_CST_LOW (t), sizeof (HOST_WIDE_INT), 1, s);
          fread (&TREE_INT_CST_HIGH (t), sizeof (HOST_WIDE_INT), 1, s);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          load_flags (t, s);
          break;
        }

      case REAL_CST:
        {
          t = make_node (code);
          load_flags (t, s);
          fread (&TREE_REAL_CST (t), sizeof (REAL_VALUE_TYPE), 1, s);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          break;
        }

      case COMPLEX_CST:
        {
          t = make_node (code);
          load_flags (t, s);
          TREE_REALPART (t) = load_tree (s, depth + 1);
          TREE_IMAGPART (t) = load_tree (s, depth + 1);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          break;
        }

      case STRING_CST:
        {
          t = make_node (code);
          load_flags (t, s);
          fread (&TREE_STRING_LENGTH (t), sizeof (int), 1, s);
          TREE_STRING_POINTER (t) = oballoc (TREE_STRING_LENGTH (t) + 9);
          fread (TREE_STRING_POINTER (t), TREE_STRING_LENGTH (t) + 1, 1, s );
          TREE_TYPE (t) = load_tree (s, depth + 1);
          break;
        }

      case FUNCTION_DECL:
        {
          tree name, full_name, asmname, type, args, d, heading, a, argdecls;
          int ellipsis = 0;

          name = load_tree (s, depth + 1);
          if (current_type_name)
            full_name = get_method_name (current_type_name, name);
          else
            full_name = name;
          asmname = load_tree (s, depth + 1);
          type = load_tree (s, depth + 1);
          args = load_tree (s, depth + 1);

          type = tree_cons (NULL_TREE, type, NULL_TREE);
          type = tree_cons (NULL_TREE, extern_id, type);
          pushlevel (0);
          clear_parm_order ();
          declare_parm_level (1);
          for (a = args; a; a = TREE_CHAIN (a))
            if (TREE_VALUE (a) == void_type_node)
              ellipsis = 1;
            else
              {
                d = build_tree_list (NULL_TREE, TREE_VALUE (a));
                d = build_tree_list (build_tree_list (d,
                                       get_unique_identifier ("parm", 0)),
                                     build_tree_list (NULL_TREE, NULL_TREE));
                push_parm_decl (d);
              }
          argdecls = get_parm_info (ellipsis);
          poplevel (0, 0, 0);

          /* Avoid a warning in case this Unit is in the same file
           * as the main program.
           */
          IDENTIFIER_LIMBO_VALUE (full_name) = NULL_TREE;

          heading = build_nt (CALL_EXPR, full_name, argdecls, NULL_TREE);
          if (!asmname || full_name == asmname)
            t = grok_directive (type, heading,
                                build_tree_list (NULL_TREE, d_external), 1);
          else
            {
              DECL_ASSEMBLER_NAME (d_asmname) = 
                build_string (IDENTIFIER_LENGTH (asmname),
                              IDENTIFIER_POINTER (asmname));
              t = grok_directive (type, heading,
                                  build_tree_list (NULL_TREE, d_asmname), 1);
            }

          /* This might be a constructor or a virtual method. */
          load_flags (t, s);

          if (current_type_name)
            {
              /* Normally, t will be ignored since grok_directive already
               * has done the job of an external function declaration.  Only
               * if this is a method (i.e. current_type_name != NULL_TREE),
               * a copy of t with `name' instead of `full_name' will go as
               * a field into the object (a RECORD_TYPE node).
               */
              t = copy_node (t);
              DECL_NAME (t) = name;
            }
          break;
        }

      case LABEL_DECL:
      case PARM_DECL:
      case RESULT_DECL:
        {
          t = make_node (code);
          DECL_NAME (t) = load_tree (s, depth + 1);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          DECL_CONTEXT (t) = current_function_decl;
          load_flags (t, s);
          break;
        }

      case FIELD_DECL:
        {
          t = make_node (code);
          load_flags (t, s);
          load_decl_flags (t, s);
          DECL_NAME (t) = load_tree (s, depth + 1);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          DECL_SIZE (t) = load_tree (s, depth + 1);
          break;
        }

      case CONST_DECL:
      case TYPE_DECL:
        {
          t = make_node (code);
          load_flags (t, s);
          DECL_NAME (t) = load_tree (s, depth + 1);
          TREE_TYPE (t) = load_tree (s, depth + 1);
          DECL_INITIAL (t) = load_tree (s, depth + 1);
          DECL_CONTEXT (t) = current_function_decl;
          break;
        }

      case VAR_DECL:
        {
          /* Re-build the tree node rather than load it */
          tree name, asmname, type;
          name = load_tree (s, depth + 1);
          asmname = load_tree (s, depth + 1);
          type = load_tree (s, depth + 1);
          if (asmname == name)
            declare_vars (build_tree_list (NULL_TREE, name),
                          type, NULL_TREE, NULL_TREE, 1);
          else
            declare_vars (build_tree_list (NULL_TREE, name),
                          type, NULL_TREE,
                          build_tree_list (asmname, extern_id), 1);
          t = lookup_name (name);
          load_flags (t, s);
          break;
        }
      #undef LOAD_MAIN_VARIANT
    }
  pop_obstacks ();

  if (flag_debug_gpi && code < LAST_AND_UNUSED_TREE_CODE)
    {
      fprintf (stderr, "loaded (%d):\n", depth);
      debug_tree (t);
    }

  gpi_contents = chainon (gpi_contents, build_tree_list (NULL_TREE, t));
  return t;
}

/*
 * Create GPI files (GNU Pascal Interface) containing precompiled
 * export Interfaces of a Unit or Module.
 *
 * In addition we create a GPM file (GNU Pascal Module) pointing 
 * to the GPI files.  This is needed when an Implementation Module
 * is looking for the Interfaces.
 */
void
create_gpi_files ()
{
  tree escan;
  FILE *gpm_file;
  char *gpm_file_name, *p;
  char gpm_header [] = "GNU Pascal Module\n";

  gpm_file_name = concat (IDENTIFIER_POINTER (current_module->name), ".gpm", "");
  for (p = gpm_file_name; *p; p++)
    *p = tolower (*p);
  gpm_file = fopen (gpm_file_name, "wa");
  if (flag_debug_gpi)
    fprintf (stderr, "storing GPM header: %s", gpm_header);
  fprintf (gpm_file, gpm_header);
  for (escan = current_module->exports; escan; escan = TREE_CHAIN (escan))
    {
      tree name = TREE_VALUE (escan);
      store_string (gpm_file, IDENTIFIER_POINTER (name));
      if (flag_debug_gpi)
        fprintf (stderr, "storing Interface name %s\n", IDENTIFIER_POINTER (name));
    }
  store_string (gpm_file, "");
  fclose (gpm_file);

  for (escan = current_module->exports; escan; escan = TREE_CHAIN (escan))
    {
      FILE *gpi_file;
      char *gpi_file_name, *p;
      tree name = TREE_VALUE (escan), nscan, iscan;
      char gpi_header [] = "GNU Pascal Unit/Module Interface\n";

      gpi_file_name = concat (IDENTIFIER_POINTER (name), ".gpi", "");
      for (p = gpi_file_name; *p; p++)
        *p = tolower (*p);
      gpi_file = fopen (gpi_file_name, "wb");
      if (flag_debug_gpi)
        fprintf (stderr, "storing GPI header: %s", gpi_header);
      fwrite (gpi_header, 33, 1, gpi_file);
      if (flag_debug_gpi)
        fprintf (stderr, "storing Module name: %s\n", IDENTIFIER_POINTER (current_module->name));
      store_string (gpi_file, IDENTIFIER_POINTER (current_module->name));
      if (flag_debug_gpi)
        fprintf (stderr, "storing Module source file name: %s\n", main_input_filename);
      store_string (gpi_file, main_input_filename);
      gpi_contents = NULL_TREE;
      /* Store names of interfaces imported by this module */
      if (flag_debug_gpi)
        fprintf (stderr, "storing names of imported Interfaces:\n");
      for (iscan = current_module->imports; iscan;
           iscan = TREE_CHAIN (iscan))
        {
          /* Store them as strings, not as tree nodes. */
          /* We don't yet want to use the gpi_contents mechanism. */
          tree iname = TREE_VALUE (iscan);
          if (!iname || TREE_CODE (iname) != IDENTIFIER_NODE)
            abort ();
          if (flag_debug_gpi)
            fprintf (stderr, "  %s\n", IDENTIFIER_POINTER (iname));
          store_string (gpi_file, IDENTIFIER_POINTER (iname));
        }
      store_string (gpi_file, "");
      store_tree (TREE_VALUE (escan), gpi_file, 0);
      for (nscan = TREE_PURPOSE (escan); nscan; nscan = TREE_CHAIN (nscan))
        if (TREE_VALUE (nscan)        /* both may be NULL_TREE */
            || TREE_PURPOSE (nscan))
          {
            if (TREE_CODE (TREE_VALUE (nscan)) != IDENTIFIER_NODE)
              warning ("Internal GPC problem: exporting non-identifier");
            else if (lookup_name (TREE_VALUE (nscan)) == NULL_TREE)
              error ("exported name `%s' undefined",
                     IDENTIFIER_POINTER (TREE_VALUE (nscan)));
            store_tree (TREE_VALUE (nscan), gpi_file, 0);   /* original name */
            store_tree (TREE_PURPOSE (nscan), gpi_file, 0);  /* renamed name */
          }
      store_tree (NULL_TREE, gpi_file, 0);
      fclose (gpi_file);
    }
}

/*
 * Open a gpi file for reading and process the header.
 * Do an AutoMake, if necessary and/or requested.
 * The file is closed via a normal fclose ().
 */
FILE *
gpi_open (name, source, depth)
     char *name;
     char *source;
     int depth;
{
  FILE *s = NULL;
  char module_name [256], module_filename [256];
  if (access (name, 4) == 0)
    s = fopen (name, "rb");
  if (s)
    {
      char gpi_loaded_string [256];
      tree import_list = NULL_TREE;

      fread (gpi_loaded_string, 33, 1, s);
      gpi_loaded_string [33] = 0;
      if (flag_debug_gpi)
        fprintf (stderr, "loaded GPI header: %s", gpi_loaded_string);
      load_string (s, module_name);
      load_string (s, module_filename);
      if (flag_debug_gpi)
        {
          fprintf (stderr, "loaded Module name: %s\n",
                           module_name);
          fprintf (stderr, "loaded Module source file name: %s\n",
                           module_filename);
          fprintf (stderr, "loaded names of imported Interfaces:\n");
        }
      do
        {
          /* Store names of imported Interfaces for later use */
          load_string (s, gpi_loaded_string);
          if (gpi_loaded_string [0])
            {
              if (flag_debug_gpi)
                fprintf (stderr, "  %s\n", gpi_loaded_string);
              import_list = chainon (import_list,
                                     build_tree_list (NULL_TREE,
                                       get_identifier (gpi_loaded_string)));
            }
        }
      while (gpi_loaded_string [0]);
      if (flag_automake && depth == 0)
        if (module_must_be_recompiled (name, module_filename))
          {
            fclose (s);
            if (flag_debug_gpi)
              fprintf (stderr, "recompiling: %s --> %s\n",
                               module_filename, name);
            compile_module (module_filename);
            /* Module has been compiled (or not).  Reload the GPI file. */
            s = gpi_open (name, source, depth + 1);
          }
        else
          {
            /* No recompilation needed.  Also check modules used by this one. */
            FILE *u;
            tree imported, i_name;
            char *u_name, *p;

            for (imported = import_list; imported;
                 imported = TREE_CHAIN (imported))
              {
                i_name = TREE_VALUE (imported);
                u_name = concat (IDENTIFIER_POINTER (i_name), ".gpi", "");
                for (p = u_name; *p; p++)
                  *p = tolower (*p);
                u = gpi_open (u_name, NULL, 0);
                if (u)
                  fclose (u);
              }
          }
    }
  else if (flag_automake && depth == 0)
    {
      /* GPI file not found.  Try to compile the module by using the
       * name given in SOURCE.  The user may omit the extension and/or
       * use capital letters in the filename.
       */
      if (source)
        {
          int result;
          strcpy (module_filename, source);  /* First, try the given name. */
          result = compile_module (module_filename);
          if (result != 0)
            {
              strcat (module_filename, ".p");  /* Next, try extension `.p' */
              result = compile_module (module_filename);
            }
          if (result != 0)
            {
              strcat (module_filename, "as");  /* Next, try extension `.pas' */
              result = compile_module (module_filename);
            }
          if (result != 0)
            {
              /* No success yet.  But the user did specify the filename
               * by a string constant.  Try decapitalized version.
               */
              char *p = module_filename;
              strcpy (module_filename, source);
              while (*p)
                {
                  *p = tolower (*p);
                  p++;
                }
              result = compile_module (module_filename);
            }
          if (result != 0)
            {
              strcat (module_filename, ".p");  /* Once more, extend with `.p' */
              result = compile_module (module_filename);
            }
          if (result != 0)
            {
              strcat (module_filename, "as");  /* Last try: `.pas' */
              compile_module (module_filename);
            }
        }
      /* SOURCE not given.  Try to derive the source file name from
       * the interface name.
       */
      else
        {
          char *p;
          strcpy (module_filename, name);
          /* Cut the extension `.gpi' */
          p = module_filename + strlen (module_filename);
          do
            p--;
          while (p >= module_filename && *p != '.');
          *p = 0;
          strcat (module_filename, ".p");  /* First, try extension `.p' */
          if (compile_module (module_filename) != 0)
            {
              strcat (module_filename, "as");  /* Now try extension `.pas' */
              compile_module (module_filename);
            }
        }
      /* Module has been compiled (or not). */
      /* Try again to load the GPI file.    */
      s = gpi_open (name, source, depth + 1);
    }
  return s;
}

/*
 * Try to load a GPI file and to extract an exported module interface.
 */
tree
load_gpi_file (name, source)
     char *name;
     char *source;
{
  FILE *gpi_file;
  char *gpi_file_name, *p;
  tree export;

  we_are_loading_a_gpi_file = 1;

  gpi_file_name = concat (name, ".gpi", "");
  for (p = gpi_file_name; *p; p++)
    *p = tolower (*p);
  gpi_file = gpi_open (gpi_file_name, source, 0);
  if (gpi_file)
    {
      tree value, purpose, name, rname;
      gpi_contents = NULL_TREE;
      value = load_tree (gpi_file, 0);
      purpose = NULL_TREE;
      do
        {
          name = load_tree (gpi_file, 0);  /* original name */
          if (name)
            {
              rname = load_tree (gpi_file, 0);  /* renamed name */
              purpose = chainon (purpose, build_tree_list (rname, name));
            }
        }
      while (name);
      fclose (gpi_file);
      export = build_tree_list (purpose, value);
    }
  else
    export = NULL_TREE;

  we_are_loading_a_gpi_file = 0;
  return export;
}

/*
 * Try to load a GPM file and to extract the interface part of this Module.
 */
int
load_gpm_file ()
{
  FILE *gpm_file;
  char *gpm_file_name, *p;
  char gpm_line [256];

  gpm_file_name = concat (IDENTIFIER_POINTER (current_module->name), ".gpm", "");
  for (p = gpm_file_name; *p; p++)
    *p = tolower (*p);
  gpm_file = fopen (gpm_file_name, "ra");

  if (!gpm_file)
    return -1;

  fread (gpm_line, 18, 1, gpm_file);
  gpm_line [18] = 0;
  if (flag_debug_gpi)
    fprintf (stderr, "loaded GPM header: %s", gpm_line);
  do
    {
      tree exported;

      load_string (gpm_file, gpm_line);
      if (gpm_line [0])
        {
          if (flag_debug_gpi)
            fprintf (stderr, "loaded Interface name: %s\n", gpm_line);
          exported = load_gpi_file (gpm_line, NULL);
          if (exported)
            {
              tree ename;
              for (ename = TREE_PURPOSE (exported); ename;
                   ename = TREE_CHAIN (ename))
                {
                  /* "Activate" all names
                   */
                  if (TREE_VALUE (ename))
                    IDENTIFIER_LOCAL_VALUE (TREE_VALUE (ename)) =
                        IDENTIFIER_LIMBO_VALUE (TREE_VALUE (ename));
                }
              export_interface (get_identifier (gpm_line),
                                TREE_PURPOSE (exported));
            }
        }
    }
  while (gpm_line [0]);
  fclose (gpm_file);
  return 0;
}



/*
 * NAME is an IDENTIFIER_NODE of the interface name.
 *
 * IMPORT_QUALIFIER:
 *     NULL_TREE if no qualifiers given.
 *     TREE_LIST:
 *       TREE_PURPOSE:
 *           NULL_TREE  -> no restricted_import_option given
 *           !NULL_TREE -> restricted_import_option (== ONLY) given
 *	 TREE_VALUE:
 *           TREE_LIST
 *             TREE_PURPOSE: imported <name> from the interface
 *             TREE_VALUE:
 *		   NULL_TREE: name has not been renamed
 *		   identifier_node: new name of the renamed <name>
 *
 * QUALIFIED_IMPORT is 0 if unqualified references are allowed;
 *                     1 if qualified references are mandatory.
 *
 * UNIT_FILENAME is the name of the source file of this Unit.
 * (Not used with Modules.)
 */
void
import_interface (interface, import_qualifier, qualified_import, unit_filename)
     tree interface;
     tree import_qualifier;
     long qualified_import;
     tree unit_filename;
{
  tree exported_name_list;
  tree exported;
  tree name;

  current_module->imports = chainon (current_module->imports,
				     build_tree_list (NULL_TREE,
						      interface));

  if (interface == standard_interface_input)
    {
      current_module->input_file_node = global_input_file_node;
      exported_name_list = build_tree_list (NULL_TREE, identifier_input);
    }
  else if (interface == standard_interface_output)
    {
      current_module->output_file_node = global_output_file_node;
      exported_name_list = build_tree_list (NULL_TREE, identifier_output);
    }
  else
    {
      for (exported = exported_interface_list;
	   exported;
	   exported = TREE_CHAIN (exported))
	if (TREE_VALUE (exported) == interface)
	  break;

      if (! exported)
        {
          char *gpi_source;
          if (unit_filename)
            gpi_source = TREE_STRING_POINTER (unit_filename);
          else
            gpi_source = NULL;
          exported = load_gpi_file (IDENTIFIER_POINTER (interface),
                                    gpi_source);
          if (exported)
            exported_interface_list = chainon (exported_interface_list,
                                               exported);
          else
	    {
	      error ("no exported interface matching `%s'",
		     IDENTIFIER_POINTER (interface));
	      return;
	    }
        }

      exported_name_list = TREE_PURPOSE (exported);
    }

  if (qualified_import)
    warning ("`qualified' not yet supported - ignored");

  /* EXPORTED is now the correct interface */

  if (import_qualifier)
    {
      tree iname;
      for (iname = TREE_VALUE (import_qualifier); iname;
           iname = TREE_CHAIN (iname))
	{
	  tree ename = exported_name_list;

	  while (ename && TREE_VALUE (ename) != TREE_PURPOSE (iname))
            ename = TREE_CHAIN (ename);

	  if (ename && TREE_VALUE (ename) == TREE_PURPOSE (iname))
            {
              tree local;
              /* Activate this name and resolve import/export renaming.
               */
              if (TREE_VALUE (iname))  /* import renaming */
                local = TREE_VALUE (iname);
              else if (TREE_PURPOSE (ename))  /* export renamimg */
                local = TREE_PURPOSE (ename);
              else  /* no renaming */
                local = TREE_VALUE (ename);
              IDENTIFIER_LOCAL_VALUE (local) =
                  IDENTIFIER_LIMBO_VALUE (TREE_VALUE (ename));
              if (TREE_READONLY (TREE_VALUE (ename)))
                TREE_READONLY (IDENTIFIER_LOCAL_VALUE (local)) = 1;
              IDENTIFIER_LIMBO_VALUE (TREE_VALUE (ename)) = NULL_TREE;
            }
          else
	    error ("interface `%s' does not export `%s'",
		   IDENTIFIER_POINTER (interface),
                   IDENTIFIER_POINTER (TREE_PURPOSE (iname)));
	}
    }

  if (import_qualifier == NULL_TREE  /* without qualifiers */
      || TREE_PURPOSE (import_qualifier) == NULL_TREE)  /* without `only' */
    {
      tree ename;
      for (ename = exported_name_list; ename; ename = TREE_CHAIN (ename))
        {
          /* "Activate" all remaining names and resolve export renaming.
           */
          if (IDENTIFIER_LIMBO_VALUE (TREE_VALUE (ename)))
            {
              tree local;
              if (TREE_PURPOSE (ename))
                local = TREE_PURPOSE (ename);
              else
                local = TREE_VALUE (ename);
              IDENTIFIER_LOCAL_VALUE (local) =
                  IDENTIFIER_LIMBO_VALUE (TREE_VALUE (ename));
              if (TREE_READONLY (TREE_VALUE (ename)))
                TREE_READONLY (IDENTIFIER_LOCAL_VALUE (local)) = 1;
              IDENTIFIER_LIMBO_VALUE (TREE_VALUE (ename)) = NULL_TREE;
            }
        }
    }
}

/*
 * Only exportable names allowed as the NAME here:
 *	  constant_name
 *	| type_name
 *	| schema_name
 *	| procedure_name
 *	| function_name
 *	| variable_name
 *	| PROTECTED variable_name
 */
tree
module_export_clause (name, renamed, protected)
     tree name;
     tree renamed;
     int  protected;
{
  tree rval = build_tree_list (renamed, name);
  TREE_READONLY (name) = protected;
  return rval;
}

tree
module_export_range (low, high)
     tree low;
     tree high;
{
  module_export_clause (build_tree_list (low, high), 0, 0);
}

/* Replace the exported TREE_LISTs denoting ranges by
 * the actual IDENTIFIER nodes.  Exported enumerated types
 * are handled here, too.
 */
void
module_repair_exported_ranges ()
{
  tree escan;
  for (escan = current_module->exports; escan; escan = TREE_CHAIN (escan))
    {
      tree nscan = TREE_PURPOSE (escan);
      while (nscan)
        {
          tree exported = NULL_TREE;
          tree meaning;

          if (TREE_VALUE (nscan)
              && TREE_CODE (TREE_VALUE (nscan)) == IDENTIFIER_NODE
              && (meaning = lookup_name (TREE_VALUE (nscan))) != NULL_TREE
              && TREE_CODE (meaning) == TYPE_DECL  /* @@@ or better always? */
              && TREE_TYPE (meaning)
              && TREE_CODE (TREE_TYPE (meaning)) == ENUMERAL_TYPE)
            {
              /* "Repair" exported enumerated types here.
               */
              tree item;
              for (item = TYPE_VALUES (TREE_TYPE (meaning)); item;
                   item = TREE_CHAIN (item))
                {
                  exported = chainon (exported,
                                      build_tree_list (NULL_TREE,
                                                       TREE_PURPOSE (item)));
                }
            }

          else if (TREE_VALUE (nscan)
                   && TREE_CODE (TREE_VALUE (nscan)) == TREE_LIST)
            {
              tree low = TREE_PURPOSE (TREE_VALUE (nscan));
              tree high = TREE_VALUE (TREE_VALUE (nscan));
              tree type = TREE_TYPE (lookup_name (low));

              if (TREE_CODE (type) != ENUMERAL_TYPE
                  || TREE_TYPE (lookup_name (high)) != type)
                error ("Internal GPC error: cannot export range");
              else
                {
                  tree item;
                  int export_it = 0;
                  for (item = TYPE_VALUES (type); item;
                       item = TREE_CHAIN (item))
                    {
                      if (TREE_PURPOSE (item) == low)
                        export_it = 1;
                      if (export_it)
                        exported = chainon (exported,
                                     build_tree_list (NULL_TREE,
                                                      TREE_PURPOSE (item)));
                      if (TREE_PURPOSE (item) == high)
                        {
                          if (! export_it)
                            error ("wrong order in exported range");
                          export_it = 0;
                        }
                    }
                }

            }
          /* Chain EXPORTED to NSCAN.
           * A TREE_LIST at NSCAN will be removed later.
           */
          if (exported)
            {
              tree new_nscan = exported;
              while (TREE_CHAIN (new_nscan))
                new_nscan = TREE_CHAIN (new_nscan);
              TREE_CHAIN (nscan) = chainon (exported, TREE_CHAIN (nscan));
              nscan = new_nscan;
            }
          nscan = TREE_CHAIN (nscan);
        }
 
      /* Remove the TREE_LISTs from the chain of exported names.
       */
      for (nscan = TREE_PURPOSE (escan); nscan; nscan = TREE_CHAIN (nscan))
        if (TREE_VALUE (nscan)
            && TREE_CODE (TREE_VALUE (nscan)) == TREE_LIST)
          TREE_VALUE (nscan) = TREE_PURPOSE (nscan) = NULL_TREE;
    }
}
